/**
 * @file capture-dc.c
 * Interface for capturing still frames from a IEEE 1394 Desktop Camera
 * @author Marko Mkel (msmakela@nic.funet.fi)
 */

/*
 * Copyright  2004 Marko Mkel
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <libraw1394/raw1394.h>
#include <libdc1394/dc1394_control.h>
#include <stdio.h>
#include "capture.h"
#include "image.h"

#if IMAGE_WIDTH != 640 || IMAGE_HEIGHT != 480 || IMAGE_DEPTH != 3
#error IMAGE_WIDTH, IMAGE_HEIGHT, IMAGE_DEPTH != 640, 480, 3
#endif

/** raw handle to the IEEE 1394 bus */
static raw1394handle_t handle;
/** libdc1394 capturing context */
static dc1394_cameracapture camera;

/** Set up the camera for capturing frames
 * @param dev	the Video1394 device descriptor name
 * @return	zero on success, nonzero on failure
 */
int
capture_init (const char* dev)
{
  int numCameras;
  int frame_rate;
  quadlet_t supported_framerates;
  nodeid_t* camera_nodes;

  if (!(handle = dc1394_create_handle (0))) {
    fputs ("Unable to acquire a raw1394 handle\n", stderr);
    return 1;
  }

  camera_nodes = dc1394_get_camera_nodes (handle, &numCameras, 0);

  if (numCameras < 1) {
    fputs ("no cameras found\n", stderr);
  cleanup1:
    dc1394_destroy_handle (handle);
    return 1;
  }

  if (dc1394_query_supported_framerates(handle, camera_nodes[0],
					FORMAT_VGA_NONCOMPRESSED,
					MODE_640x480_RGB,
					&supported_framerates) !=
      DC1394_SUCCESS) {
    fputs ("dc1394_query_supported_framerates() failed\n", stderr);
    goto cleanup1;
  }

  if (supported_framerates & (1U << 31))
    frame_rate = FRAMERATE_1_875;
  else if (supported_framerates & (1U << 30))
    frame_rate = FRAMERATE_3_75;
  else if (supported_framerates & (1U << 29))
    frame_rate = FRAMERATE_7_5;
  else if (supported_framerates & (1U << 28))
    frame_rate = FRAMERATE_15;
  else if (supported_framerates & (1U << 27))
    frame_rate = FRAMERATE_30;
  else if (supported_framerates & (1U << 26))
    frame_rate = FRAMERATE_60;
  else {
    fputs ("unsupported frame rate\n", stderr);
    goto cleanup1;
  }

  if (dc1394_dma_setup_capture (handle, camera_nodes[0],
				0, /* channel */ 
				FORMAT_VGA_NONCOMPRESSED,
				MODE_640x480_RGB,
				SPEED_400,
				frame_rate,
				4 /* num_dma_buffers*/,
				0 /* do_extra_buffering */,
				1 /* drop_frames*/,
				dev, &camera) != DC1394_SUCCESS) {
    fputs ("unable to setup camera\n", stderr);
  cleanup2:
    dc1394_dma_release_camera (handle, &camera);
    goto cleanup1;
  }

  if (dc1394_start_iso_transmission (handle, camera.node)
      != DC1394_SUCCESS) {
    fputs ("unable to start camera iso transmission\n", stderr);
    dc1394_dma_unlisten (handle, &camera);
    goto cleanup2;
  }

  return 0;
}

/** Clean up the camera after capturing frames */
void
capture_fini (void)
{
  if (dc1394_stop_iso_transmission (handle, camera.node))
    fputs ("could not stop the camera?\n", stderr);
  dc1394_dma_unlisten (handle, &camera);
  dc1394_dma_release_camera (handle, &camera);
  dc1394_destroy_handle (handle);
  handle = 0;
}

/** Capture a frame
 * @param com	(output) image comment, NUL terminated
 * @param size	maximum length of the comment buffer, including terminating NUL
 * @return	pointer to frame data, or NULL on failure
 */
const char*
capture_do (char* com,
	    size_t size)
{
  /* capture the metadata */
  if (size) {
    dc1394_camerainfo info;
    quadlet_t ver;
    unsigned i, j;
    /** Metadata available from libdc1394 */
    static const struct metadata {
      int (*func) (raw1394handle_t, nodeid_t, unsigned*);
      const char* label;
    } settings[] = {
      { dc1394_get_brightness, "Brightness" },
      { dc1394_get_exposure, "Exposure" },
      { dc1394_get_sharpness, "Sharpness" },
      { dc1394_get_hue, "Hue" },
      { dc1394_get_saturation, "Saturation" },
      { dc1394_get_gamma, "Gamma" },
      { dc1394_get_shutter, "Shutter" },
      { dc1394_get_gain, "Gain" },
      { dc1394_get_iris, "Iris" },
      { dc1394_get_focus, "Focus" },
      { dc1394_get_trigger_mode, "Trigger Mode" },
      { dc1394_get_zoom, "Zoom" },
      { dc1394_get_pan, "Pan" },
      { dc1394_get_tilt, "Tilt" },
      { dc1394_get_optical_filter, "Optical Filter" },
      { dc1394_get_capture_size, "Capture Size" },
      { dc1394_get_capture_quality, "Capture Quality" },
      { 0, 0 }
    };
    static const struct metadata* setting;
    int s;

    if (dc1394_get_camera_info (handle, camera.node, &info) ==
	DC1394_SUCCESS) {
      s = snprintf (com, size, "Vendor: %s\nModel: %s\n",
		    info.vendor, info.model);
      if (s > 0 && s < size)
	size -= s, com += s;
    }

    if (dc1394_get_sw_version(handle, camera.node, &ver) == DC1394_SUCCESS) {
      const char* verstr;
      /* mapping taken from coriander 1.0.0-pre3, update_frames.c */
      switch (ver) {
      case 0x100: verstr = "1.04"; break;
      case 0x101: verstr = "1.20"; break;
      case 0x102: verstr = "1.30"; break;
      case 0x114: verstr = "Point Grey 114"; break;
      default:
	s = snprintf (com, size, "IIDC Version: 0x%x\n", ver);
	goto got_version;
      }
      s = snprintf (com, size, "IIDC Version: %s\n", verstr);
    got_version:
      if (s > 0 && s < size)
	size -= s, com += s;
    }

    for (setting = settings; setting->func; setting++) {
      if ((*setting->func) (handle, camera.node, &i) == DC1394_SUCCESS) {
	s = snprintf (com, size, "%s: %u\n", setting->label, i);
	if (s <= 0 || s >= size)
	  break;
	size -= s, com += s;
      }
    }

    if (dc1394_get_white_balance (handle, camera.node, &i, &j) ==
	DC1394_SUCCESS) {
      s = snprintf (com, size, "White Balance: BU=%u RV=%u\n", i, j);
      if (s > 0 && s < size)
	size -= s, com += s;
    }

    if (dc1394_get_temperature (handle, camera.node, &i, &j) ==
	DC1394_SUCCESS) {
      s = snprintf (com, size, "Temperature: goal=%u current=%u\n", i, j);
      if (s > 0 && s < size)
	size -= s, com += s;
    }
  }

  return dc1394_dma_single_capture (&camera) == DC1394_SUCCESS
    ? (const char*) camera.capture_buffer
    : 0;  
}

/** Done with a frame */
void
capture_done (void)
{
  dc1394_dma_done_with_buffer (&camera);
}
