/**
 * @file jpeg.c
 * JPEG compression
 * @author Marko Mkel (msmakela@nic.funet.fi)
 */

#include <stdio.h>
#include "jpeg.h"
#include "image.h"

#if IMAGE_DEPTH != 3
#error IMAGE_DEPTH != 3
#endif

/** Compress the image.
 * @param src		uncompressed image, in RGB colour space,
 *			width * depth * 3 bytes
 * @param width		image width
 * @param height	image height
 * @param quality	JPEG compression quality (1..100, libjpeg default=75)
 * @param comment	image comment (optional)
 * @param length	length of image comment, in bytes
 * @param cinfo		libjpeg compression context
 * @return		zero on success, nonzero on failure
 */
int
jpeg_output (const char* src,
	     unsigned width, unsigned height,
	     int quality,
	     const void* comment, int length,
	     struct jpeg_compress_struct* cinfo)
{
  cinfo->image_width = width;
  cinfo->image_height = height;
  cinfo->input_components = IMAGE_DEPTH;
  cinfo->in_color_space = JCS_RGB;
  jpeg_set_defaults (cinfo);

  jpeg_set_quality (cinfo, quality, FALSE);
  cinfo->optimize_coding = TRUE;

  jpeg_start_compress (cinfo, TRUE);

  if (comment && length)
    jpeg_write_marker (cinfo, JPEG_COM, comment, length);

  while (cinfo->next_scanline < cinfo->image_height) {
    JSAMPROW row_pointer =
      (JSAMPROW) &src[cinfo->next_scanline * width * IMAGE_DEPTH];
    jpeg_write_scanlines (cinfo, &row_pointer, 1);
  }

  (void) jpeg_finish_compress (cinfo);
  return 0;
}
