/**
 * @file sender.c
 * Captures images and sends them to a child process
 * @author Marko Mkel (msmakela@nic.funet.fi)
 */

/*
 * Copyright  2004 Marko Mkel
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* protocol:
 * sender invokes receiver
 *  receiver sends a command "JPEG width height quality"
 *  sender responds by sending JPEG data in blocks
 *  - each block is prepended by a 16-bit big-endian number indicating length
 *  - end of image is signalled by length 0
 *  - error is signalled by length 0 immediately
 *  - after completing request or reporting error, server returns to main loop
 * - on SIGCHLD or SIGPIPE, sender should terminate with nonzero status
 * - must test if SIGKILL or SIGINT leave zombie processes
 * - surrounding shell script will restart sender,
 *   e.g., in case of communications failure
 */

#define _POSIX_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#include <sys/types.h>
#include <unistd.h>
#include "capture.h"
#include "lenscorr.h"
#include "image.h"
#include "jpeg.h"

/** libjpeg error manager */
struct my_error_mgr
{
  /** "public" fields */
  struct jpeg_error_mgr pub;
  /** for returning to caller */
  jmp_buf setjmp_buffer;
};

/**
 * Custom error exit handler
 * @param cinfo		the libjpeg context
 */
static void
my_error_exit (struct jpeg_common_struct* cinfo)
{
  /* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
  struct my_error_mgr* myerr = (struct my_error_mgr*) cinfo->err;

  /* Always display the message. */
  /* We could postpone this until after returning, if we chose. */
  (*cinfo->err->output_message) (cinfo);

  /* Return control to the setjmp point */
  longjmp (myerr->setjmp_buffer, 1);
}

/** output to child (image files) */
static FILE* child_out;
/** input from child (commands) */
static FILE* child_in;
/** output buffer for image data */
static JOCTET child_outbuf[16384];

/** Initialize or terminate the child_output buffer
 * @param cinfo	libjpeg compression context
 */
static void
child_out_init (struct jpeg_compress_struct* cinfo)
{
  cinfo->dest->next_output_byte = child_outbuf;
  cinfo->dest->free_in_buffer = sizeof child_outbuf;
}

/** Terminate or terminate the child_output buffer
 * @param cinfo	libjpeg compression context
 */
static void
child_out_term (struct jpeg_compress_struct* cinfo)
{
  size_t size = cinfo->dest->next_output_byte - child_outbuf;
  putc (size >> 8, child_out), putc (size, child_out);
  fwrite (child_outbuf, 1, size, child_out);
}

/** Empty the child_output buffer
 * @param cinfo	libjpeg compression context
 */
static boolean
child_out_empty (struct jpeg_compress_struct* cinfo)
{
  putc ((sizeof child_outbuf) >> 8, child_out);
  putc (sizeof child_outbuf, child_out);
  fwrite (child_outbuf, 1, sizeof child_outbuf, child_out);
  cinfo->dest->next_output_byte = child_outbuf;
  cinfo->dest->free_in_buffer = sizeof child_outbuf;
  return TRUE;
}

/** Fork a child process and set up communications
 * @param argv	command line arguments for the child process
 * @return	zero on success, nonzero on failure
 */
static int
create_child (char** argv)
{
  /** pipes for the image output and the command input */
  int to_child[2], from_child[2];
  pid_t pid;

  if (pipe (to_child)) {
    perror ("pipe");
    return 1;
  }
  if (pipe (from_child)) {
    perror ("pipe");
    close (to_child[0]), close (to_child[1]);
    return 1;
  }
  if (!(pid = fork ())) {
    /* child process */

    if (dup2 (to_child[0], STDIN_FILENO) < 0 ||
	close (to_child[1]) < 0 ||
	close (from_child[0]) < 0 ||
	dup2 (from_child[1], STDOUT_FILENO) < 0) {
      perror ("dup2 or close");
      exit (127);
    }
    if (to_child[0] != STDIN_FILENO)
      close (to_child[0]);
    if (from_child[1] != STDOUT_FILENO)
      close (from_child[1]);
    execvp (*argv, argv);
    perror (*argv);
    exit (126);
  }
  else if (pid < 0) {
    perror ("fork");
    close (to_child[0]), close (to_child[1]);
    close (from_child[0]), close (from_child[1]);
    return 2;
  }
  close (to_child[0]), close (from_child[1]);

  if (!(child_in = fdopen (from_child[0], "r")) ||
      !(child_out = fdopen (to_child[1], "wb"))) {
    perror ("fdopen");
    if (child_in)
      fclose (child_in);
    else
      close (from_child[0]);
    close (to_child[1]);
    return 3;
  }

  return 0;
}

/** Main program
 * @param argc	number of command line arguments
 * @param argv	command line arguments
 */
int
main (int argc, char** argv)
{
  const char* src;
  char* dst = 0;
  size_t dst_size = 0;
  unsigned width, height, quality;
  struct jpeg_compress_struct cinfo;
  struct my_error_mgr jerr;
  /** end of file marker */
  static const char eof[2] = { 0, 0 };
  /** image comment (camera settings etc.) */
  static char comment[2048];

  if (argc < 3) {
    fprintf (stderr, "usage: %s /dev/video1394/0 client [args]\n", *argv);
    return 1;
  }

  jerr.pub.error_exit = my_error_exit;
  cinfo.err = jpeg_std_error (&jerr.pub);

  /* Establish the setjmp return context for my_error_exit to use. */
  if (setjmp (jerr.setjmp_buffer)) {
    return 2; /* failure in libjpeg */
  }

  jpeg_create_compress (&cinfo);

  if (capture_init (argv[1]))
    return 2;

  if (create_child (argv + 2)) {
    capture_fini ();
    free (dst);
    jpeg_destroy_compress (&cinfo);
    return 1;
  }

  jpeg_stdio_dest (&cinfo, 0);
  cinfo.dest->init_destination = child_out_init;
  cinfo.dest->term_destination = child_out_term;
  cinfo.dest->empty_output_buffer = child_out_empty;

  for (;;) {
    size_t size;
    int c;

    if (fflush (child_out)) {
      perror ("fflush");
      break;
    }

    if (3 != fscanf (child_in, "JPEG %u%u%u", &width, &height, &quality) ||
	!width || !height) {
      static const char err_unknown[] =
	"\0\0\"JPEG width height quality;\" expected\n";
      if (feof (child_in) || ferror (child_in)) {
	perror ("read from child");
	break;
      }
      fwrite (err_unknown, (sizeof err_unknown) - 1, 1, child_out);
      fputs ("invalid request\n", stderr);
      /* skip anything until next line */
      do
	c = getc (child_in);
      while (c != EOF && c != '\n');
      continue;
    }
    /* read until end of line */
    do
      c = getc (child_in);
    while (c != EOF && c != '\n');

    fprintf (stderr, "JPEG %u %u %u...", width, height, quality);
    fflush (stderr);

    size = width * height * IMAGE_DEPTH;
    if (size > dst_size) {
      char* new_dst = realloc (dst, size);
      if (!new_dst) {
	fwrite (eof, sizeof eof, 1, child_out);
	fprintf (child_out, "JPEG %u %u %u: not enough memory\n",
		 width, height, quality);
	continue;
      }
      dst = new_dst, dst_size = size;
    }

    if (!(src = capture_do (comment, sizeof comment))) {
      fwrite (eof, sizeof eof, 1, child_out);
      fprintf (child_out, "JPEG %u %u %u: capture failed\n",
	       width, height, quality);
      continue;
    }

    lens_correct (src, dst, width, height);
    capture_done ();
    jpeg_output (dst, width, height, quality,
		 comment, strlen (comment), &cinfo);
    fwrite (eof, sizeof eof, 1, child_out);
    fputs ("JPEG ok\n", child_out);

    fputs ("\b\b\b   \n", stderr);
  }
  capture_fini ();
  fclose (child_in), fclose (child_out);
  free (dst);
  jpeg_destroy_compress (&cinfo);
  return 0;
}
